/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <ctype.h>
#include <mintbind.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "general.h"

#include "clipbrd.h"
#include "clips.h"
#include "manager.h"
#include "protocols.h"
#include "utils.h"

typedef struct xattr
{
   uint16_t   mode; 			 /* File mode, statements for */
								 /* - possible filetypes	  */
								 /* - special bits			  */
								 /* - access rights 		  */
   int32_t	  index;			 /* File number 			  */
   uint16_t   dev;				 /* Device number			  */
   uint16_t   rdev; 			 /* Actual device			  */
								 /* (e.g. for BIOS files)	  */
								 /* in MagiC however reserved */
   uint16_t   nlink;			 /* Number of links 		  */
   uint16_t   uid;				 /* User number 			  */
   uint16_t   gid;				 /* Group number			  */
   int32_t	  size; 			 /* File length 			  */
   int32_t	  blksize, nblocks;  /* Blocksize/occupied blocks */
   uint16_t   mtime, mdate; 	 /* Date of last modification */
   uint16_t   atime, adate; 	 /* Date of the last access   */
   uint16_t   ctime, cdate; 	 /* Creation date			  */
   uint16_t   attr; 			 /* TOS file attributes 	  */
   uint16_t   reserved2;		 /* Reserviert				  */
   int32_t	  reserved3[2]; 	 /* Reserviert				  */
} XATTR;

static void process_clipboard (GEMCLIP *manager, USERCONF *conf);
static void create_clipbrd    (char *clipbrd, const char *avserver, int os);

/*
 * This function is called each time GEMClip polls the clipboard, which is every second or so.
 * We must always verify that the clipboard path hasn't been modified by another application with
 * scrap_write(). So we take some precautions and write the path if neccessary. NOTE: is it possible
 * to get the actual boot drive? NOTE: why do we have a userconf clipbrd and a manager clipbrd?
 */
void handle_clipboard(GEMCLIP *manager, USERCONF *conf)
{
	char *p     = NULL;
	int newpath = FALSE;

	if (manager->suspend == FALSE)
	{
		/* Is there a clipboard path define in the system? Don't just assume the path returned is right. */
		if (scrp_read(manager->clipbrd) != 0)
		{
			if (add_trailing_backslash(manager->clipbrd, LBUFFER) != NULL)
			{
				process_clipboard(manager, conf);
			}
			else { // Path too long?
				strcpy(manager->clipbrd, conf->clipbrd);
			}
		}
		else
		{
			/* Check for some possible env variables, but only if the user does not want to override
			   the system setting with the user conf setting. NOTE: HypView messes up the clipboard.
			   Should try and fix that. Obviously, we don't process the clipboard, since it is empty. */
			if (conf->override == FALSE)
			{
				shel_envrn(&p, "CLIPBRD=");

				if (p == NULL) {
					shel_envrn(&p, "SCRAPDIR=");
				}

				/* Check that we have a path and that it is correctly formatted. */
				if (p != NULL)
				{
					if (add_trailing_backslash(p, LBUFFER) != NULL)
					{
						strcpy(manager->clipbrd, p);
						scrp_write(manager->clipbrd);
					}
					else
					{
						strcpy(manager->clipbrd, conf->clipbrd); // Last resort solution, since nothing works.
						newpath = TRUE;
					}
				}
				else
				{
					strcpy(manager->clipbrd, conf->clipbrd);
					newpath = TRUE;
				}
			}
			else
			{
				strcpy(manager->clipbrd, conf->clipbrd);
				newpath = TRUE;
			}

			/* Create the clipboard directory (or use an already existing dir that hasn't been
			   set as default by a scrp_write() call) and write the new path. */
			if (newpath == TRUE) {
				create_clipbrd(manager->clipbrd, manager->avserver, manager->system);
			}
		}
	}
}

/*
 * Main function for processing the clipboard. This is the sequence of actions:
 * - Look for a scrap.txt file.
 * - Check if that file was modified.
 * - If it was, reorder clips.
 * - Copy scrap.txt to clip00.
 * - Update Manager window, if opened, to reflect the change (or do it anyway?).
 *
 * <== Parameters:
 *   - clipbrd path
 *   - overall structure
 */
static void process_clipboard(GEMCLIP *manager, USERCONF *conf)
{
	char path[LBUFFER] = {0};
	int scraphdl = -1;
	int dirhdl   = -1;

	/* We suppose, if this fails, that the directory does not exist. */
	if ((dirhdl = open_clipbrd(manager->clipbrd)) >= 0)
	{
		snprintf(path, sizeof path, "%s%s", manager->clipbrd, "scrap.txt");

		/* Is there a scrap.txt file? */
		if ((scraphdl = Fopen(path, 0)) >= 0)
		{
			if (manager->clipn == NOSCRAP)
			{
				manager->clipn = SCRAP;
				update_objects(manager);
			}
		}

		/* Update clips then redraw the portions of the Manager that were modified, 
		   if the Manager is open, or the icon. */
		if (update_manager(scraphdl, dirhdl, manager, conf) == TRUE)
		{
			if (manager->window->icon == TRUE) {
				update_icon_preview(manager);
			}
			else if (manager->window->open == TRUE)
			{
				update_clips_infoline(manager);
				set_listbox_order(manager, conf);
				update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
				redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
				redraw_object(manager->window->handle, manager->tree, MANAGER_PREVIEW_BOX);
			}
		}

		/* If this was disabled, re-enable all toolbar icons. */
		if (scraphdl >= 0)
		{
			if (manager->tree[MANAGER_CLIP_ICON].ob_state & OS_DISABLED)
			{
				manager->tree[MANAGER_SAVE].ob_state &= ~OS_DISABLED;
				manager->tree[MANAGER_VIEW].ob_state &= ~OS_DISABLED;
				manager->tree[MANAGER_CLIP_ICON].ob_state &= ~OS_DISABLED;
				
				if (manager->window->open == TRUE)
				{
					redraw_object(manager->window->handle, manager->tree, MANAGER_VIEW);
					redraw_object(manager->window->handle, manager->tree, MANAGER_SAVE);
					redraw_object(manager->window->handle, manager->tree, MANAGER_CLIP_ICON);
				}
				manager->clipn = SCRAP;
			}
		}

		if ((scraphdl < 0) && (manager->nclips > 0)) // No SCRAP.TXT but some clips.
		{
			int emptied = FALSE; // We need to know if the preview boxes have been emptied in this call.

			if (*manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_ptext != '\0')
			{
				*manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_ptext = '\0';
				emptied = TRUE;
			}

			/* Clear preview box if no clip is selected. */
			if (manager->selected == 0)
			{
				for (int i = MANAGER_PREV1; i <= MANAGER_PREV6; i++) {
					*manager->tree[i].ob_spec.tedinfo->te_ptext = '\0';
				}
			}

			if ((manager->window->open == TRUE) && (emptied == TRUE))
			{
				redraw_object(manager->window->handle, manager->tree, MANAGER_CLIPBOARD);
				redraw_object(manager->window->handle, manager->tree, MANAGER_PREVIEW_BOX);
			}
		}
		else if ((scraphdl < 0) && (manager->nclips == 0)) { // Completely empty.
			clear_manager(manager);
		}
		
		if (scraphdl >= 0) {
			Fclose(scraphdl);
		}
		
		Dclosedir(dirhdl);
	}
 	else if ((dirhdl == -34) || (dirhdl == -33)) { // Create clipbrd directory. There's nothing to process.
		create_clipbrd(manager->clipbrd, manager->avserver, manager->system);
	}
}

/*
 * Open the clipboard directory. Return the directory handle or a negative GEMDOS error code.
 */
int open_clipbrd(const char *clipbrd)
{
	char path[LBUFFER] = {0};
	char *p    = NULL;

	/* We must remove the trailing backlash for the Dopendir() call. But since we need
	   the unmodified path, we copy it 1st. Both arrays are the same size. */
	strcpy(path, clipbrd);

	if ((p = (strrchr(path, '\\'))) != NULL) {
		*p = '\0';
	}

	return Dopendir(path, 0);
}

/*
 * Either create the clipboard directory (or use an already existing one that hasn't been set as
 * default by a scrp_write() call) and write the new path. We take care not to recreate that dir if
 * the clipboard path is already valid. NOTE: should rewrite this. Too complicated.
 */
static void create_clipbrd(char *clipbrd, const char *avserver, int os)
{
	char path[LBUFFER] = {0};
	int err = -1;

	if (scrp_read(path) != 0)
	{
		err = Dcreate(path);

		if ((err == E_OK) || (err == EACCDN)) // Either no error or directory already exists.
		{
			update_desktop_window(path, avserver, os);
		}
		else
		{
			err = Dcreate(clipbrd);

			if ((err == E_OK) || (err == EACCDN))
			{
				scrp_write(clipbrd);
				update_desktop_window(clipbrd, avserver, os);
			}
			else {
				form_alert(1, FA_ERROR "[Error creating clipboard|directory.][OK]");
			}
		}
	}
	else
	{
		err = Dcreate(clipbrd);

		if ((err == E_OK) || (err == EACCDN))
		{
			scrp_write(clipbrd);
			update_desktop_window(clipbrd, avserver, os);
		}
		else {
			form_alert(1, FA_ERROR "[Error creating clipboard|directory.|scrp_read() returned 0.][OK]");
		}
	}
}

/*
 * Delete all clip files.
 */
void empty_clipboard(const char *clipbrd)
{
	char buffer[LBUFFER] = {0};
	char path[LBUFFER]   = {0};
	int dirhdl           = -1;
	long xr	             = 0;
	XATTR xattr;

	if ((dirhdl = open_clipbrd(clipbrd)) >= 0)
	{
		while (Dxreaddir(sizeof buffer, dirhdl, buffer, &xattr, &xr) != ENMFIL)
		{
			if ((buffer[4] == 'c') && (buffer[5] == 'l') && (buffer[6] == 'i') && (buffer[7] == 'p'))
			{
				snprintf(path, sizeof path, "%s%s", clipbrd, &buffer[4]);

				if (Fdelete(path) != E_OK) {
					form_alert(1, FA_ERROR "[Error deleting file|in empty_clipboard().][OK]");
				}
			}
		}
	}
}

/*
 * Delete no longer needed clips after reducing the clipboard size.
 */
void resize_clipboard(GEMCLIP *manager, int newsize)
{
	char buffer[LBUFFER] = {0};
	char path[LBUFFER]   = {0};
	int dirhdl           = -1;
	int clipn            = 0;
	long xr	             = 0;
	XATTR xattr;

	if ((dirhdl = open_clipbrd(manager->clipbrd)) >= 0)
	{
		while (Dxreaddir(sizeof buffer, dirhdl, buffer, &xattr, &xr) != ENMFIL)
		{
			clipn = strtol(&buffer[8], NULL, 10);
			
			/* We take care of deleting only our own files. */
			if ((clipn > newsize) &&
				((buffer[4] == 'c') && (buffer[5] == 'l') && (buffer[6] == 'i') && (buffer[7] == 'p')))
			{
				snprintf(path, sizeof path, "%s%s", manager->clipbrd, &buffer[4]);

				if (Fdelete(path) != E_OK) {
					form_alert(1, FA_ERROR "[Error deleting file|in empty_clipboard().][OK]");
				}
			}
		}
		/* Force an update of the Manager without creating a new clip. */
    	manager->clips[SCRAP]->time.year = 0;
		send_local_message(SC_CHANGED, 0x0002, 0, 0, 0, 0);
	}
}
