/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <mintbind.h>
#include <mint/ostruct.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "general.h"

#include "clips.h"
#include "info.h"
#include "protocols.h"
#include "utils.h"

static int find_next_selected  (GEMCLIP *manager);
static int find_prev_selected(GEMCLIP *manager);

/*
 * Create the Information window dialog. Defaults to scrap.txt if no clip are selected.
 */
int open_info_dialog(WINDOW *window, GEMCLIP *manager)
{
	static int reopen = FALSE;
	int attribs  = NAME | CLOSER | MOVER | SMALLER | INFO;
	OBJECT *tree = get_gem_tree(INFOS);
	GRECT rect;
	GRECT wrect;

	if (reopen == FALSE) // First opening.
	{
		/* Calculate the coordinates to open the dialog in the centre of the screen. */
		wind_get_grect(0, WF_WORKXYWH, &wrect);
	
		rect.g_x = wrect.g_x + ((wrect.g_w - tree->ob_width)  / 2);
		rect.g_y = wrect.g_y + ((wrect.g_h - tree->ob_height) / 2);
		rect.g_w = tree->ob_width;
		rect.g_h = tree->ob_height;

		wind_calc_grect(WC_BORDER, attribs, &rect, &wrect);
	}
	else { // Re-opening.
		wrect = manager->infrect;
	}

	if ((window = init_window(window)) != NULL)
	{
		/* Create and open the window. */
		if ((window->handle = wind_create_grect(attribs, &wrect)) != 0)
		{
			/* Enable/disable 'Next' button according to the # of selected clips. */
			if (manager->selected < 2) {
				tree[INFOS_NEXT].ob_state |= OS_DISABLED;
			}
			else {
				tree[INFOS_NEXT].ob_state &= ~OS_DISABLED;
			}

			tree[INFOS_PREV].ob_state |= OS_DISABLED; // 'Next' always DISABLED.

			wind_set_str(window->handle, WF_NAME, get_gstring(STRINGS, STRINGS_INFO_NAME));

			manager->clipn = find_first_selected(manager);
			process_info(manager, window, tree);

			/* Open Information window. */
			if (wind_open_grect(window->handle, &wrect) != 0)
			{
				window->open = TRUE;
				reopen = TRUE;
			}
		}
	}

	return window->open;
}

/*
 * Process button and keyboard actions on the Info dialog.
 */
void handle_info_dialog(GEMCLIP *manager, WINDOW *window, OBJECT *tree, EVMULT_OUT *evmult_out, short object, USERCONF *conf, BUBBLES *bubbles)
{
	if ((evmult_out->emo_mbutton & LEFT_BUTTON) && (tree[object].ob_state & OS_DISABLED)) {
		sound_bell();
	}
	else
	{
		if (object == 0) // Key pressed.
		{
			switch (evmult_out->emo_kreturn)
			{
			case SC_HELP:
				if (manager->aeslang == AESLANG_GERMAN) {
					call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPD.HYP ", "Benutzung", manager->system);
				}
				else {
					call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPE.HYP ", "Usage", manager->system);
				}
				break;
			case SC_UNDO:
			case SC_RETURN:
			case SC_NMPAD_ENTER:
				tree[INFOS_CLOSE].ob_state &= ~OS_SELECTED;
				send_local_message(WM_CLOSED, window->handle, 0, 0, 0, 0);
				break;
			}
		}
		else if ( ! (evmult_out->emo_mbutton & RIGHT_BUTTON))
		{
			static int sel = 1;
		
			switch (object) // Button clicked.
			{
			case INFOS_CLOSE:
				tree[object].ob_state &= ~OS_SELECTED;
				send_local_message(WM_CLOSED, window->handle, 0, 0, 0, 0);
				break;
			case INFOS_PREV:
				if (find_prev_selected(manager) == TRUE)
				{
					sel--;
					process_info(manager, window, tree);
	
					if (sel <= 1)
					{
						tree[object].ob_state |= OS_DISABLED;
					}
	
					if (sel < manager->selected)
					{
						tree[INFOS_NEXT].ob_state &= ~OS_DISABLED;
					}
					tree[object].ob_state &= ~OS_SELECTED;
					redraw_object(window->handle, tree, ROOT);
				}
				break;
			case INFOS_NEXT:
				if (find_next_selected(manager) == TRUE)
				{
					sel++;
					process_info(manager, window, tree);
	
					if (sel >= manager->selected)
					{
						tree[object].ob_state |= OS_DISABLED;
					}
	
					if (sel > 1)
					{
						tree[INFOS_PREV].ob_state &= ~OS_DISABLED;
					}
					tree[object].ob_state &= ~OS_SELECTED;
					redraw_object(window->handle, tree, ROOT);
				}
				break;
			case INFOS_HELP:
				if (manager->aeslang == AESLANG_GERMAN) {
					call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPD.HYP ", "Benutzung", manager->system);
				}
				else {
					call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPE.HYP ", "Usage", manager->system);
				}
				break;
			}
		}
		else
		{
			switch (object)
			{
			case INFOS_CLOSE:
				display_bubble_mouse(bubbles, ABTOKAY, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case INFOS_PREV:
				display_bubble_mouse(bubbles, INFPREV, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case INFOS_NEXT:
				display_bubble_mouse(bubbles, INFNEXT, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case INFOS_HELP:
				display_bubble_mouse(bubbles, INFHELP, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			}
		}
	}
}

/*
 * Process the content and count approximative number of chars, words, spaces and lines.
 */
void process_info(GEMCLIP *manager, WINDOW *win, OBJECT *tree)
{
	char buffer[sizeof ((WINDOW*)0)->info] = {0};
	char *content  = NULL;
	char *str      = NULL;
	char *p        = NULL;
	int file       = 0;
	int lines      = 0;
	int words      = 0;
	int txtlen     = 0;
	int clipn      = manager->clipn;
	size_t size    = 0;
	OBJECT *strings = get_gem_tree(STRINGS);

	if (clipn == SCRAP)
	{
		wind_set_str(win->handle, WF_INFO, strings[STRINGS_INFO_NAME].ob_spec.free_string); // Reuse title.
		strcpy(win->info, strings[STRINGS_INFO_NAME].ob_spec.free_string);
		strcpy(tree[INFOS_NAME].ob_spec.tedinfo->te_ptext, "SCRAP.TXT");
		snprintf(buffer, sizeof buffer, "%s%s", manager->clipbrd, "SCRAP.TXT");
	}
	else // Clip file.
	{
		/* Build infoline. */
		snprintf(buffer,
				 sizeof buffer,
				 "%s %d %s %d %s",
				 strings[STRINGS_CLIPN].ob_spec.free_string,
				 clipn,
				 strings[STRINGS_OUT_OF].ob_spec.free_string,
				 manager->selected,
				 strings[STRINGS_SLT_CLIPS].ob_spec.free_string);

		wind_set_str(win->handle, WF_INFO, buffer);
		strcpy(win->info, buffer);

		/* Write clip name. */
		str = tree[INFOS_NAME].ob_spec.tedinfo->te_ptext;
		txtlen = tree[INFOS_NAME].ob_spec.tedinfo->te_txtlen;
		snprintf(str, txtlen, "%s %d", strings[STRINGS_CLIPN].ob_spec.free_string, clipn/*&manager->clips[clipn]->name[4]*/);

		write_clippath(buffer, sizeof buffer, manager->clipbrd, clipn);
	}
	
	if ((file = Fopen(buffer, S_READ)) >= 0)
	{
		size = Fseek(0, file, SEEK_END); // Get file size. We don't need the return value.

		if ((content = xmalloc(size)) != NULL)
		{
			if (Fseek(0, file, SEEK_SET) >= 0) // Rewind.
			{
				if (Fread(file, size, content) >= 0)
				{
					content[size] = '\0';
				
					/* Process file for size information. */
					for (p = content; *p; p++)
					{
						switch (*p)
						{
						case '\n':
							/* Don't count several lines for succesive newlines. */
							if (*(p + 1) != '\n') {
								lines++;
							}
							break;
						case '\r':
							lines++;
							break;
						case ' ':
							/* Don't count several words for succesive spaces. */
							if (*(p + 1) != ' ') {
								words++;
							}
							break;
						case '\t':
							/* Don't count several words for succesive tabs. */
							if (*(p + 1) != '\t') {
								words++;
							}
							break;
						}
					}
				}
				else {
					form_alert(1, FA_ERROR "[Error reading file|in process_info().][OK]");
				}
			}
			else {
				form_alert(1, FA_ERROR "[Error seeking file|in process_info().][OK]");
			}

			free(content);
		}
		Fclose(file);
	}
	else {
		form_alert(1, FA_ERROR "[Error opening file|in process_info().][OK]");
	}

	if (lines == 0) {
		lines++;
	}

	/* Write file size information. */
	txtlen = tree[INFOS_CHARS_INC].ob_spec.tedinfo->te_txtlen;
	snprintf(tree[INFOS_CHARS_INC].ob_spec.tedinfo->te_ptext, txtlen, "%ld", size);
	snprintf(tree[INFOS_CHARS_EXC].ob_spec.tedinfo->te_ptext, txtlen, "%ld", size - words);
	snprintf(tree[INFOS_LINES].ob_spec.tedinfo->te_ptext, txtlen, "%d", lines);
	snprintf(tree[INFOS_WORDS].ob_spec.tedinfo->te_ptext, txtlen, "%d", words + 1);

	/* Write date and time info, then path. */
	str = tree[INFOS_DATE].ob_spec.tedinfo->te_ptext;
	txtlen = tree[INFOS_DATE].ob_spec.tedinfo->te_txtlen;

	switch (manager->dformat)
	{
	case 0:
		snprintf(str,
				 txtlen,
				 "%02d%c%02d%c%d",
				 manager->clips[clipn]->time.month, manager->datesep,
				 manager->clips[clipn]->time.day, manager->datesep, manager->clips[clipn]->time.year + 1980);
		break;
	case 256:
		snprintf(str,
				 txtlen,
				 "%02d%c%02d%c%d",
				 manager->clips[clipn]->time.day, manager->datesep,
				 manager->clips[clipn]->time.month, manager->datesep, manager->clips[clipn]->time.year + 1980);
		break;
	case 512:
		snprintf(str,
				 txtlen,
				 "%d%c%02d%c%02d",
				 manager->clips[clipn]->time.year + 1980, manager->datesep,
				 manager->clips[clipn]->time.month, manager->datesep, manager->clips[clipn]->time.day);
		break;
	case 768:
		snprintf(str,
				 txtlen,
				 "%d%c%02d%c%02d",
				 manager->clips[clipn]->time.year + 1980, manager->datesep,
				 manager->clips[clipn]->time.day, manager->datesep, manager->clips[clipn]->time.month);
		break;
	}

	str = tree[INFOS_TIME].ob_spec.tedinfo->te_ptext;
	txtlen = tree[INFOS_TIME].ob_spec.tedinfo->te_txtlen;
	snprintf(str,
			 txtlen,
			 "%02d:%02d:%02d",
			 manager->clips[clipn]->time.hour,
			 manager->clips[clipn]->time.minute,
			 manager->clips[clipn]->time.second * 2);

	str = tree[INFOS_PATH].ob_spec.tedinfo->te_ptext;
	txtlen = tree[INFOS_PATH].ob_spec.tedinfo->te_txtlen;
	snprintf(str, txtlen, "%s", manager->clipbrd);
}

/*
 * Find the next selected clip in the array. Returns TRUE if next clip found, FALSE otherwise.
 */
static int find_next_selected(GEMCLIP *manager)
{
	int found = FALSE;

	if (manager->selected > 1)
	{
		manager->clipn++;

		while (found == FALSE)
		{
			if (manager->clipn <= manager->nclips)
			{
				if (manager->clips[manager->clipn]->select == TRUE)	{
					found = TRUE;
				}
				else {
					manager->clipn++;
				}
			}
		}
	}

	return found;
}

/*
 * Find the next selected clip in the array. Returns TRUE if next clip found, FALSE otherwise.
 */
static int find_prev_selected(GEMCLIP *manager)
{
	int found = FALSE;

	manager->clipn--;

	while (found == FALSE)
	{
		if (manager->clipn > 1)
		{
			if (manager->clips[manager->clipn]->select == TRUE)	{
				found = TRUE;
			}
			else {
				manager->clipn--;
			}
		}
	}

	return found;
}
