/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <mint/osbind.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "general.h"

#include "protocols.h"
#include "utils.h"
#include "av.h"

/*
 * Set up and open a virtual workstation.
 * ==> Return the handle of the workstation or 0 on error.
 */
VdiHdl open_virtual_workst(short workin[], short workout[])
{
	short foo	  = 0;
	VdiHdl aeshdl = 0; // Must contain O or 1 according to the GEMlib doc.
	VdiHdl apphdl = 0;

	aeshdl = graf_handle(&foo, &foo, &foo, &foo);
	apphdl = aeshdl;
	v_opnvwk(workin, &apphdl, workout);

	if (apphdl == 0) {
		form_alert(1, FA_ERROR "[Error while opening virtual|workstation.][OK]");
	}

	return apphdl;
}

/*
 * Search for a key in a file and read the associated value (organized as a 'KEY = VALUE' pair).
 * Reset file access position to the start of the file, to avoid values being skipped when the
 * config file structure changes.
 * <== Parameters:
	 - a file descriptor;
	 - a key;
	 - a buffer LBUFFER bytes long to write the value into;
	 - the size of the buffer.
 * ==> Return the value or NULL.
 */
char *get_value(FILE *file, const char *key, char value[], size_t size)
{
	char *p	= NULL;
	char *ctrl = NULL; // Exit loop when NULL;
	char line[LBUFFER] = {0};

	/* Read each line of the file, skipping over # and spaces. */
	do
	{
		ctrl = fgets(line, LBUFFER, file);

		if ((ctrl != NULL) && (isalpha(*line)))
		{
			/* Search for the key and value. */
			if ((strstr(line, key) != NULL) && ((p = strchr(line, '=')) != NULL))
			{
				/* line may not end with a newline if file's been tampered with. */
				line[strlen(line) - 1] = '\0';

				p++; // Skip over the '='.

				while (isblank(*p)) {
					p++;
				}

				if (*p != '\0')
				{
					/* Take care of empty strings. */
					if ((*p == '"') && (*(p + 1) == '"'))
					{
						value[0] = '\0';
						ctrl = NULL;
					}
					else if (size > 0)
					{
						strcpy(value, p);
						ctrl = NULL;
					}
				}
				else {
					ctrl = NULL;
				}
			}
		}
	} while (ctrl != NULL);

	rewind(file);

	return (value[0] != '\0') ? value : NULL;
}

/*
 * Handle a window dialog. Find on which button the event occurred then process
 * key input into editable fields.
 * ==> Return the index of the object which triggered the event or 0.
 */
int do_window_form(OBJECT *tree, WINDOW *window, EVMULT_OUT *evmult_out, short event)
{
	short object  = -1;
	short editobj = 0;
	short newobj  = 0;
	short retobj  = 0;
	short keyout  = 0;
	short idx	  = 0;

	if (event & MU_BUTTON)
	{
		object = objc_find(tree, ROOT, MAX_DEPTH, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y);

		if (object != -1)
		{
			/* If it's a right mouse click, don't process the event, only return the object. */
			if ( ! (evmult_out->emo_mbutton & RIGHT_BUTTON))
			{
				if (((form_wbutton(tree, object, 1, &newobj, window->handle)) == 1) && (newobj > 0))
				{
					if (tree[editobj].ob_flags & OF_EDITABLE) { // Is editobj an edit field?
						objc_wedit(tree, editobj, 0, 0, ED_END, window->handle);
					}

					objc_wedit(tree, newobj, 0, &idx, ED_INIT, window->handle);
					editobj = newobj;
				}
				else {
					retobj = object;							// Return the object index.
				}
			}
			else {
				retobj = object;								// Return the object index.
			}
		}
	}
	else if (event & MU_KEYBD)
	{
		if (form_wkeybd(tree, editobj, editobj, evmult_out->emo_kreturn, &newobj, &keyout, window->handle))
		{
			if (newobj != editobj)
			{
				if (tree[editobj].ob_flags & OF_EDITABLE) { // Is editobj an edit field?
					objc_wedit(tree, editobj, 0, 0, ED_END, window->handle);
				}

				editobj = newobj;
			}

			if (keyout) {									// Process the key.
				objc_wedit(tree, editobj, keyout, &idx, ED_CHAR, window->handle);
			}
		}
	}

	if (newobj != editobj && newobj != 0)
	{
		if (editobj != 0) { 								// Stop editing.
			objc_wedit(tree, editobj, 0, &idx, ED_END, window->handle);
		}

		retobj = newobj;
	}

	return retobj;
}

/*
 * Update the infoline after a MU_M1 message.
 */
void handle_mouse_rect(GEMCLIP *manager, WINDOW *win, OBJECT *tree, BUBBLES *bubbles, EVMULT_IN *evmult_in, EVMULT_OUT *evmult_out)
{
	short object          = 0;
	static short previous = 0;

	if (evmult_in->emi_m1leave == MO_ENTER)
	{	
		object = objc_find(tree, ROOT, MAX_DEPTH, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y);
	
		/* If object not the same as previous one. */
		if (object != previous)
		{
			/* Skip un-needed objects. Also, if the button doesn't have an
			associated help message, write the default config dialog name. */
			if ((tree[object].ob_flags & OF_SELECTABLE) ||
				(tree[object].ob_flags & OF_EXIT)		||
				(tree[object].ob_flags & OF_TOUCHEXIT)	||
				(tree[object].ob_flags & OF_EDITABLE))
			{
				update_infoline(manager, win, tree, bubbles, object);
			
				/* Set new grect then ask for a MU_M1 message when the grect is left. */
				objc_offset(tree, object, &evmult_in->emi_m1.g_x, &evmult_in->emi_m1.g_y);
				evmult_in->emi_m1.g_w = tree[object].ob_width;
				evmult_in->emi_m1.g_h = tree[object].ob_height;
				evmult_in->emi_m1leave = MO_LEAVE;
			}
			else { // Only rewrite infoline.
				update_infoline(manager, win, tree, bubbles, object);
			}
		}
	}
	else if (evmult_in->emi_m1leave == MO_LEAVE)
	{
		wind_get_grect(win->handle, WF_WORKXYWH, &evmult_in->emi_m1);
		evmult_in->emi_m1leave = MO_ENTER;
		wind_set_str(win->handle, WF_INFO, win->info);
	}

	previous = object;
}

/*
 * Handle selection in popup menus. Makes use of the AES 4.1 SCROLL_LISTBOX option.
 * <== Parameters:
	 - handle of window
	 - object tree containing the object which activated the function
	 - object tree containing the popup menu itself
	 - object index of the button in the first tree
	 - scrolling mode wanted
	 - buffer to copy the G_STRING in
	 - size of the buffer
 * ==> Return the selected menu item or 0.
 */
short handle_popup(short handle, OBJECT *tree, OBJECT *popup, short object, int mode, char *buffer, size_t size)
{
	GRECT rect;
	MENU menu;
	MENU mdata;

	menu.mn_tree   = popup; // The menu (G_BOX + G_STRING's).
	menu.mn_menu   = ROOT;	// Parent is always 0.
	menu.mn_item   = 1; 	// The last used item.
	menu.mn_scroll = mode;	// SCROLL_NO, SCROLL_YES or SCROLL_LISTBOX.

	if (objc_offset(tree, object, &rect.g_x, &rect.g_y) != 0)
	{
		/* Adapt the x and y position to make the popup open just below the button
		and not on top of it. Cosmetic adjustment (NOTE: hardcoded since computations with
		te_thickness didn't do anything here). */
		if (menu_popup(&menu, rect.g_x, rect.g_y + tree[object].ob_height + 3, &mdata) == 1)
		{
			/* Copy label from the calling object tree, skipping over the two leading spaces. */
			snprintf(buffer, size, "%s", &mdata.mn_tree[mdata.mn_item].ob_spec.free_string[1]);

			/* If the object calling the function has a label, copy G_STRING into it. */
			switch (tree[object].ob_type & 0xFF) // Mask off type.
			{
			/* If ob_spec is a pointer to a TEDINFO structure. */
			case G_TEXT:
			case G_FTEXT:
			case G_BOXTEXT:
			case G_FBOXTEXT:
				if (tree[object].ob_spec.tedinfo->te_ptext != NULL)
				{
					snprintf(tree[object].ob_spec.tedinfo->te_ptext, size, "%s", buffer);
				}
				break;
			/* If ob_spec is a pointer to a free string. */
			case G_STRING:
				if (tree[object].ob_spec.free_string != NULL)
				{
					size_t strsize = strlen(mdata.mn_tree[mdata.mn_item].ob_spec.free_string) + 1;
					snprintf(tree[object].ob_spec.free_string, strsize, "%s", buffer);
				}
				break;
			case G_BUTTON:
				if (tree[object].ob_spec.free_string != NULL)
				{
					int len = size; // Field must have type int.
					snprintf(tree[object].ob_spec.free_string, size, "%-*s", len - 2, &buffer[2]);
				}
				break;
			}

			/* Redraw the object calling the function and check/uncheck menu item, if needed. */
			rect.g_w = tree[object].ob_width;
			rect.g_h = tree[object].ob_height;
			objc_wdraw(tree, ROOT, MAX_DEPTH, &rect, handle);

			uncheck_prev_item(popup);
			popup[mdata.mn_item].ob_state |= OS_CHECKED;
		}
		else
		{
			buffer = NULL;
			mdata.mn_item = 0;
		}
	}
	else {
		buffer = NULL;
	}

	return mdata.mn_item;
}

/*
 * Uncheck a previously checked menu item in a popup menu.
 */
void uncheck_prev_item(OBJECT *menu)
{
	int n = 0;

	/* Count # of items in the popup. */
	while ( ! (menu[n++].ob_flags & OF_LASTOB))
		;

	for (int i = 1; i < n; i++)
	{
		if (menu[i].ob_state & OS_CHECKED) {
			menu_icheck(menu, i, UNCHECK);
		}
	}
}

/*
 * Resize an AES 4 popup menu after adding/deleting an item.
 * <== Parameters:
 *	 - a pointer to the menu
 *	 - the number of items
 *	 - the width in chars of the longest string.
 */
void resize_popup(OBJECT *tree, int nitems, size_t width)
{
	short apgout = 0;
	short foo	 = 0;

	/* Get font size and adapt width. */
	if (appl_getinfo(AES_LARGEFONT, &apgout, &foo, &foo, &foo) == 0) {
		width = (width + 1) * apgout;
	}
	else {
		width = (width + 1) * 8; // Fall back to standard VDI font size.
	}

	/* Resize menu. */
	tree->ob_height = tree[1].ob_height * nitems;
	tree->ob_width = width;
}

/*
 * Redraw a window dialog after a WM_REDRAW message.
 */
void draw_window_form(OBJECT *tree, short handle, GRECT *clip)
{
	GRECT wrect;
	GRECT rect;

	wind_get_grect(handle, WF_WORKXYWH, &wrect);
	tree->ob_x = wrect.g_x;
	tree->ob_y = wrect.g_y;

	wind_update(BEG_UPDATE);
	graf_mouse(M_OFF, NULL);
	wind_get_grect(handle, WF_FIRSTXYWH, &rect); // Get first rect.

	while (rect.g_w && rect.g_h)
	{
		if (rc_intersect(clip, &rect) != 0)
		{
			objc_wdraw(tree, ROOT, MAX_DEPTH, &rect, handle);
		}
		wind_get_grect(handle, WF_NEXTXYWH, &rect); // Next rect.
	}

	graf_mouse(M_ON, NULL);
	wind_update(END_UPDATE);
}

/*
 * Draw an icon in an iconified window. The object tree may contain any number of icons.
 * <== Parameters:
 *	 - object tree containing icons;
 *	 - index in object tree of icon to draw;
 *	 - width in pixels of icon to draw;
 *	 - height...;
 *	 - handle of window in which to draw the icon.
 */
void draw_icon(OBJECT *tree, short icon, short width, short height, short handle)
{
	int nb = ROOT; // # of icons in the tree.
	GRECT rect;

	/* Count # of icons in the tree. */
	while ( ! (tree[nb++].ob_flags & OF_LASTOB))
		;

	/* Hide other icons. i is set to 1st icon. nb = # of icons + 1. */
	for (int i = 1; i < nb; i++)
	{
		if (i != icon)
		{
			tree[i].ob_flags |= OF_HIDETREE;
		}
	}

	tree[icon].ob_flags &= ~OF_HIDETREE;

	/* Position background (root object), centre icon horizontally and vertically
	then redraw window. */
	wind_get_grect(handle, WF_WORKXYWH, &rect);
	tree[ROOT].ob_x = rect.g_x;
	tree[ROOT].ob_y = rect.g_y;
	tree[icon].ob_x = (rect.g_w / 2) - (width / 2);
	tree[icon].ob_y = (rect.g_h / 2) - (height / 2);
	objc_wdraw(tree, ROOT, 1, &rect, handle);
}

/*
 * Redraw a GEM object.
 */
void redraw_object(short handle, OBJECT *tree, short object)
{
	GRECT rect;

	objc_offset(tree, object, &rect.g_x, &rect.g_y);
	rect.g_w = tree[object].ob_width;
	rect.g_h = tree[object].ob_height;
	objc_wdraw(tree, ROOT, MAX_DEPTH, &rect, handle);
}

/*
 * Change the state of a GEM object and redraw it.
 */
void change_state(short handle, OBJECT *tree, short object, int state)
{
	GRECT rect;

	objc_offset(tree, object, &rect.g_x, &rect.g_y);
	rect.g_x -= 2;
	rect.g_y -= 2;
	rect.g_w = tree[object].ob_width + 2;
	rect.g_h = tree[object].ob_height + 2;
	objc_wchange(tree, object, state, &rect, handle);
}

/*
 * Clear an editable field and set the cursor to the first position in it.
 */
void clear_editfield(OBJECT *tree, short object, short handle)
{
	short foo = 0;

	*tree[object].ob_spec.tedinfo->te_ptext = '\0';
	redraw_object(handle, tree, object);
	objc_wedit(tree, object, 0, &foo, ED_INIT, handle);
}

/*
 * Add a trailing backslash to the string and NULL-terminate it.
 * <== Parameters:
 *   - the string to modify
 *   - the total size of the array holding the string
 * ==> Return the string received.
 */
char *add_trailing_backslash(char *str, size_t size)
{
	size_t len = strlen(str);

	/* Avoid writing past the end of the buffer. */
	if ((str[len - 1] != '\\') && (len < (size - 2)))
	{
		str[len] = '\\';
		str[len + 1] = '\0';
	}

	return str;
}

/*
 * Remove whitespace from str.
 */
void remove_whitespace(char *str)
{
	char *p = str;
	char buffer[strlen(str) + 1];

	/* Completely skip whitespace from beginning, for a neater display. */
	while (isspace(*p))	{
		p++;
	}

	*buffer = '\0';
	strncat(buffer, p, sizeof buffer);
	p = buffer;

	while (*p++ != '\0')
	{
		if ((*p == '\r') || (*p == '\n') || (*p == '\t') || (*p == '\v') || (*p == '\f')) {
			*p = ' ';
		}
	}

	if ((p = strrchr(buffer, '\n')) != NULL) {
		*p = '\0';
	}

	snprintf(str, sizeof buffer, "%s", buffer);
}

/*
 * Return a pointer to an object tree from a GEM resource file or NULL on error.
 */
OBJECT *get_gem_tree(short ntree)
{
	OBJECT *tree = NULL;

	if ((rsrc_gaddr(R_TREE, ntree, &tree)) == 0)
	{
		form_alert(1, FA_ERROR "[Error getting object tree.][OK]");
		tree = NULL;
	}
	return tree;
}

/*
 * Write a string into a G_BOXTEXT, G_STRING or G_FBOXTEXT string from a numeric value. For the former type,
 * If the value is less than 2 digits long, write 2 leading spaces (cosmetic correction).
 */
void write_popup_boxtext(OBJECT *tree, short object, int value)
{
	switch (tree[object].ob_type & 0xFF) // Mask off type.
	{
	case G_BOXTEXT: // Popup menu. 1 space char at the left.
		if (value > 9) {
			snprintf(tree[object].ob_spec.tedinfo->te_ptext, tree[object].ob_spec.tedinfo->te_txtlen, " %d", value);
		}
		else {
			snprintf(tree[object].ob_spec.tedinfo->te_ptext, tree[object].ob_spec.tedinfo->te_txtlen, "  %d", value);
		}
		break;
	case G_FBOXTEXT: // Editable field. No space char except for values below 9.
		if (value > 9) {
			snprintf(tree[object].ob_spec.tedinfo->te_ptext, tree[object].ob_spec.tedinfo->te_txtlen, "%d", value);
		}
		else {
			snprintf(tree[object].ob_spec.tedinfo->te_ptext, tree[object].ob_spec.tedinfo->te_txtlen, " %d", value);
		}
		break;
	default:
		break;
	}
}

/*
 * Return a pointer to a GEM free/alert string or NULL on error.
 */
char *get_free_string(short object)
{
	char *string = NULL;

	if (rsrc_gaddr(R_STRING, object, &string) == 0)
	{
		form_alert(1, FA_ERROR "[Error getting alert string.][OK]");
		string = NULL;
	}

	return string;
}

/*
 * Get a G_STRING from a dialog box.
 * <== Parameters: a tree index (the dialog box) and an object index (the G_STRING).
 * ==> Return a pointer to the G_STRING.
 */
char *get_gstring(short ntree, short object)
{
	OBJECT *tree = NULL;

	if ((rsrc_gaddr(R_TREE, ntree, &tree)) == 0)
	{
		form_alert(1, FA_ERROR "[Error getting object tree.][OK]");
		tree[object].ob_spec.free_string = NULL;
	}

	return tree[object].ob_spec.free_string;
}

/*
 * Get a numeric value from an object label.
 */
long get_number_from_label(OBJECT *tree, short object)
{
	long value = 0;

	switch (tree[object].ob_type & 0xFF) // Mask off type.
	{
	/* If ob_spec is a pointer to a TEDINFO structure. */
	case G_TEXT:
	case G_FTEXT:
	case G_BOXTEXT:
	case G_FBOXTEXT:
		value = strtol(tree[object].ob_spec.tedinfo->te_ptext, NULL, 10);
		break;
	/* If ob_spec is a pointer to a free string. */
	case G_BUTTON:
	case G_STRING:
		value = strtol(tree[object].ob_spec.free_string, NULL, 10);
		break;
	default:
		break;
	}

	return value;
}

/*
 * Slightly more practical form_alert(). Accepts two strings.
 */
void form_alert_string(int button, const char *icon, const char *s1, const char *s2)
{
	char buffer[62];

	snprintf(buffer, sizeof buffer, "%s[%s|%s][OK]", icon, s1, s2);
	form_alert(button, buffer);
}

/*
 * Get the ASCII value from a scancode.
 */
unsigned char get_ascii_value(int kreturn, int kmeta)
{
	struct keytab {
		char *unshift;
		char *shift;
		char *caps;
	};

	unsigned char key = 0;
	struct keytab *keytab;

	keytab = Keytbl((char*) - 1, (char*) - 1, (char*) - 1);

	if (kmeta) { // Shift key was on.
		key = keytab->shift[kreturn >> 8];
	}
	else {
		key = keytab->unshift[kreturn >> 8];
	}

	return key;
}

/*
 * More practical appl_write() call. Only used for auto-sent messages, hence the name.
 */
void send_local_message(short message, short msg3, short msg4, short msg5, short msg6, short msg7)
{
	short msg[8] = {0, 0, 0, 0, 0, 0, 0, 0};

	msg[0] = message;
	msg[1] = _AESapid;
	msg[3] = msg3;
	msg[4] = msg4;
	msg[5] = msg5;
	msg[6] = msg6;
	msg[7] = msg7;
	appl_write(msg[1], 16, msg);
}

/*
 * Auto-send a redraw message for the whole window workarea.
 */
void send_redraw_message(short handle)
{
	GRECT rect;

	wind_get_grect(handle, WF_WORKXYWH, &rect);
	send_local_message(WM_REDRAW, handle, rect.g_x, rect.g_y, rect.g_w, rect.g_h);
}

/*
 * Tell the AES a the content of a drive has changed and it must update any opened windows.
 * This is not as good as the AV solution, since it updates *any* opened window from this drive.
 */
void broadcast_window_draw(char *path)
{
	short msg[8] = {0, 0, 0, 0, 0, 0, 0, 0};
(void)(path);
	msg[0] = SH_WDRAW;
	msg[1] = _AESapid;
	msg[3] = 2;

	shel_write(SHW_BROADCAST, 0, 0, msg, NULL);
}

/*
 * Notify all applications that the content of the clipboard was modified.
 */
void broadcast_sc_changed(void)
{
	short msg[8] = {0, 0, 0, 0, 0, 0, 0, 0};

	msg[0] = SC_CHANGED;
	msg[1] = _AESapid;
	msg[3] = 0x0002;

	shel_write(SHW_BROADCAST, 0, 0, msg, NULL);
}

/*
 * Display help guide. Called from a dialog box.
 * <== Parameters:
 *	 - handle of window
 *	 - pointer to object tree containing the Help button
 *	 - index of object with the Help label (might be 0 if called from the menu)
 *	 - AV server name  (can't be const, may be modified)
 *	 - guide name
 *	 - page title in the guide
 */
void call_help(short handle, OBJECT *tree, short object, char *avserver, const char *guide, const char *page, int os)
{
	short id	 = -1;
	short state  = 0;
	int mode     = MX_GLOBAL | MX_PREFTTRAM;
	short msg[8] = {VA_START, _AESapid, 0, 0, 0, 0, 0, 0};
	char *cmmd	 = NULL;
	char buffer[MBUFFER] = {'x'}; // Workaround for Pexec() eating the 1st char.

	if (object != 0) // Will be 0 if called from the menu.
	{
		state  =  OS_OUTLINED;
		state &= ~OS_SELECTED;
		change_state(handle, tree, object, state);
	}

	/* If the AV server is TeraDesk, send the message directly to the HYP viewer. */
	if (strcmp(avserver, "DESKTOP ") == 0)
	{
		if ((id = appl_find("hyp_view")) < 0) {
			id = appl_find("ST-GUIDE");
		}

		if (id <= 0) // App not yet running.
		{
			shel_envrn(&cmmd, "STGUIDE=");
			snprintf(&buffer[1], sizeof buffer, "%s%s", guide, page);
			shel_write(SWM_LAUNCH, 0, CL_NORMAL, cmmd, buffer); // Launch viewer on specific page.
		}
	}
	else // Other AV server.
	{
		id = appl_find(avserver);
	}


	if (id >= 0)
	{
		int size = 0;

		if ((size = snprintf(buffer, sizeof buffer, "%s%s", guide, page)) > 0)
		{
			if (os == MINT) {
				cmmd = (char*)Mxalloc(size + 1, mode);
			}
			else if (os == MAGIC) {
				cmmd = (char*)Mxalloc(size + 1, (int)mode & 0xFF);
			}
	
			if (cmmd != NULL)
			{
				strcpy(cmmd, buffer);
	
				*(char **)(&msg[3]) = cmmd;
				appl_write(id, 16, msg);
			}
		}
	}
	else
	{
		/* The accessory may have been launched before the desktop. Try again
		   to establish communication with it then recall this function. */
		if (init_av_protokoll(os, avserver) == TRUE) {
			call_help(handle, tree, object, avserver, guide, page, os);
		}
	}
}

/*
 * Get the coordinates of the window dialog and set the GRECT event accordingly.
 */
void set_mouse_rect(short handle, EVMULT_IN *evmult_in)
{
	GRECT rect = {-1, -1, 0, 0};

	wind_get_grect(handle, WF_WORKXYWH, &rect);

	/* Set the new rect then ask for a message when it is entered. */
	evmult_in->emi_m1 = rect;
	evmult_in->emi_m1leave = MO_ENTER;
}

/*
 * Simple wrapper for malloc(). Displays a standard GEM error dialog if malloc fails.
 * Returns a pointer to allocated space or NULL on error.
 */
void *xmalloc(size_t size)
{
	void *mem = NULL;

	if (size <= 0) {
		form_alert(1, FA_ERROR "[Size zero in xmalloc()!][OK]");
	}
	else if ((mem = malloc(size)) == NULL) {
		form_error(FERR_LOWMEM);
	}

	return mem;
}

/*
 *
 */
WINDOW *init_window(WINDOW *window)
{
	for (int i = 1; i < 9; window->workin[i++] = 1) ;

	/* Open virtual workstation. */
	window->workin[0]  = Getrez() + 2;
	window->workin[10] = 2;

	if ((window->wks = open_virtual_workst(window->workin, window->workout)) == 0)
	{
		window = NULL;
	}
	else
	{
		window->handle   = -1;
		window->wchar    = 0;
		window->hchar    = 0;
		window->wcell    = 0;
		window->hcell    = 0;
		window->icon     = FALSE;
		window->open     = FALSE;
		window->strindex = 1;
	   *window->info     = '\0';
	}

	return window;
}

/*
 * Sound the system bell.
 */
void sound_bell(void)
{
	putchar('\a');
	fflush(stdout);
}
